#ifndef BINARYSTREAMS_H
#define BINARYSTREAMS_H

#include <memory>

// The obinarystream class maintains an internal memory buffer and allows the
// user to conveniently dump objects sequentially as binary data into the
// buffer.  The buffer grows automatically as needed, so there is no need to
// precalculate the length of the data.
class obinarystream
{
private:
	char* _data;
	char* _currentPtr;
	int _size;
	int _capacity;

private:
	void growBuffer()
	{
		_capacity *= 2;
		char* newData = new char[_capacity];
		memcpy(newData, _data, _size);
		delete[] _data;
		_data = newData;
		_currentPtr = &_data[_size];
	}

public:
	obinarystream()
	{
		_capacity = 16;
		_size = 0;
		_data = new char[_capacity];
		_currentPtr = _data;
	}
	
	~obinarystream()
	{
		delete[] _data;
	}

	// Return a pointer to the internal buffer, which can then be passed to
	// the memory manager, or fstream::write(), or a similar function.
	char* data() const
	{
		return _data;
	}
	
	// Return the number of bytes that have been written to the buffer.
	int size() const
	{
		return _size;
	}

	template <typename Elem>
	friend obinarystream& operator<<(obinarystream& stream, const Elem& e);
};

// By overloading operator<<, the obinarystream can be used like the other
// standard iostreams.  Using a templated operator, the user can pass any kind
// of object to the stream, and its raw binary contents will be copied into the
// buffer.  Then the current pointer is advanced so that more elements can be
// written.
template <typename Elem>
obinarystream& operator<<(obinarystream& stream, const Elem& e)
{
	while(stream._size + sizeof(Elem) > stream._capacity)
		stream.growBuffer();

	stream._size += sizeof(Elem);
	memcpy(stream._currentPtr, &e, sizeof(Elem));
	stream._currentPtr += sizeof(Elem);

	return stream;
}

// The ibinarystream class maintains an internal memory buffer of the specified
// size (passed to the constructor).  Data can be loaded into the buffer by
// accessing the pointer returned by data(), and then operator>> can be used to
// read data sequentially from the buffer.
//
// For brevity, no checking is done to verify that you don't read beyond the
// end of the buffer.
class ibinarystream
{
private:
	char* _data;
	char* _currentPtr;
	int _size;

public:
	// Allocate a buffer with the specified size, in bytes.
	ibinarystream(int sz)
	{
		_size = sz;
		_data = new char[_size];
		_currentPtr = _data;
	}
	
	~ibinarystream()
	{
		delete[] _data;
	}

	// Return a pointer to the internal buffer, which can then be passed to
	// the memory manager, or fstream::read(), or a similar function.
	char* data() const
	{
		return _data;
	}
	
	// Return the number of bytes occupied by the buffer.
	int size() const
	{
		return _size;
	}

	template <typename Elem>
	friend ibinarystream& operator>>(ibinarystream& stream, Elem& e);
};

// By overloading operator>>, the ibinarystream can be used like the other
// standard iostreams.  Using a templated operator, the user can pass any kind
// of object to the stream, and the memory occupied by the object will be filld
// with the binary data at the current position in the buffer.  Then the
// current pointer is advanced so that more elements can be read.
template <typename Elem>
ibinarystream& operator>>(ibinarystream& stream, Elem& e)
{
	memcpy(&e, stream._currentPtr, sizeof(Elem));
	stream._currentPtr += sizeof(Elem);

	return stream;
}

#endif // BINARYSTREAMS_H
